# measpy/ni.py
#
# -------------------------------------------------
# Data acquisition with National Instrument devices
# -------------------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

import nidaqmx
import nidaqmx.constants as niconst

from ._tools import siglist_to_array, t_min

import numpy as np

from datetime import datetime
from time import sleep

def _n_to_ain(n):
    return 'ai'+str(n-1)
def _n_to_aon(n):
    return 'ao'+str(n-1)

def _callback(task_handle, every_n_samples_event_type,
                number_of_samples, callback_data):
    print('Every N Samples callback invoked.')

    return 0

def ni_run_measurement(M):
    """
    Runs a measurement defined in the object of
    the class measpy.measurement.Measurement given
    as argument.

    Once the data acquisition process is terminated,
    the measurement object given in argument contains
    a property in_sig consisting of a list of signals.

    :param M: The measurement object that defines the measurement properties
    :type M: measpy.measurement.Measurement

    :return: Nothing, the measurement passed as argument is modified in place.

    """
    system = nidaqmx.system.System.local()
    nsamps = int(round(M.dur*M.fs))

    if M.device_type!='ni':
        print("Warning: deviceType != 'ni'. Changing to 'ni'.")
        M.device_type='ni'
    if M.in_device=='':
        print("Warning: no output device specified, changing to "+system.devices[0].name)
        M.in_device=system.devices[0].name

    if hasattr(M, 'in_range'):
        if M.in_range == None:
            inr = False
        else :
            inr = True
    else:
        inr = False
    if not(inr):
        val = nidaqmx.system.device.Device(M.in_device).ai_voltage_rngs[-1]
        print("Warning: no input range specified, changing to the max value of "+M.in_device+" -> "+str(val))
        M.in_range = list(val for b in M.in_map)

    if hasattr(M, 'out_sig') and M.out_sig!=None:
        outx = siglist_to_array(M.out_sig)
        tmin = t_min(M.out_sig)
        if M.out_device=='':
            print("Warning: no output device specified, changing to "+system.devices[0].name)
            M.out_device=system.devices[0].name
        if hasattr(M, 'out_range'):
            if M.out_range == None:
                outr = False
            else :
                outr = True
        else:
            outr = False
        if not(outr):
            val = nidaqmx.system.device.Device(M.out_device).ao_voltage_rngs[-1]
            print("Warning: no output range specified, changing to the max value of "+M.out_device+" -> "+str(val))
            M.out_range = list(val for b in M.out_map)
    else:
        tmin = 0

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Set up the read tasks
    if M.in_sig!=None:
        intask = nidaqmx.Task(new_task_name="in") # read task
        for i,n in enumerate(M.in_map):
            print(_n_to_ain(n))
            intask.ai_channels.add_ai_voltage_chan(
                physical_channel=M.in_device + "/" + _n_to_ain(n),
                terminal_config=niconst.TerminalConfiguration.DEFAULT,
                min_val=-M.in_range[i], max_val=M.in_range[i],
                units=niconst.VoltageUnits.VOLTS)

        intask.timing.cfg_samp_clk_timing(
            rate=M.fs,
            sample_mode=niconst.AcquisitionType.CONTINUOUS,
            samps_per_chan=nsamps)
        
        for i,iepeval in enumerate(M.in_iepe):
           if iepeval:
                intask.ai_channels[i].ai_excit_val = 0.002
                intask.ai_channels[i].ai_coupling = niconst.Coupling.AC

    # Set up the write tasks
    if M.out_sig!=None:
        outtask = nidaqmx.Task(new_task_name="out") # write task

        # Set up the write tasks, use the sample clock of the Analog input if possible
        for i,n in enumerate(M.out_map):   
            outtask.ao_channels.add_ao_voltage_chan(
                physical_channel=M.out_device + "/" + _n_to_aon(n), 
                min_val=-M.out_range[i], max_val=M.out_range[i],
                units=niconst.VoltageUnits.VOLTS)
      
        if M.in_device.startswith('myDAQ'):
            # If the device is a myDAQ card, we keep most default values
            # The myDAQ devices are set up separately because
            # there is no error messages when setting up properties
            # that are not supported, and the acquisition then fails
            print("This is a NI my DAQ device")
            outtask.timing.cfg_samp_clk_timing(
                rate=M.fs,
                sample_mode=niconst.AcquisitionType.CONTINUOUS,
                samps_per_chan=nsamps)
        else:
            try:
                # We first try to use analog input sample clock as output clock
                outtask.timing.cfg_samp_clk_timing(
                    rate=M.fs,
                    source="/" + M.in_device + "/ai/SampleClock", #"OnboardClock",
                    sample_mode=niconst.AcquisitionType.CONTINUOUS,
                    samps_per_chan=nsamps)
                print('Use of /'+ M.in_device + '/ai/SampleClock as output clock : success !')
            except:
                # If it fails, use defaults
                # Then the in/out are not synchronized
                # There is hence the possibility to use one analog input
                # to do the in/out sync (io_sync=input channel number)
                print("Error when choosing \""+"/" + M.in_device + "/ai/SampleClock\" as clock source, let's try \"OnboardClock\" ")
                outtask.timing.cfg_samp_clk_timing(
                    rate=M.fs,
                    sample_mode=niconst.AcquisitionType.CONTINUOUS,
                    samps_per_chan=nsamps)

        if len(M.out_map)==1:
            outtask.write(outx[:,0], auto_start=False)
        else:
            outtask.write(outx.T, auto_start=False)

        outtask.start() # Start the write task first, waiting for the analog input sample clock

    if M.in_sig!=None:
        y = intask.read(nsamps,timeout=M.dur+10) # Start the read task
        intask.close()
    else:
        sleep(M.dur+10)

    if M.out_sig!=None:
        outtask.close()

    if M.in_sig!=None:
        y=np.array(y).T
        if len(M.in_map)==1:
            M.in_sig[0].raw = y
            M.in_sig[0].t0 = tmin
        else:
            for i,s in enumerate(M.in_sig):
                s.raw = y[:,i]
                s.t0 = tmin

def ni_run_synced_measurement(M,in_chan=0,out_chan=0,added_time=1):
    """
    Before running a measurement, added_time second of silence
    is added at the begining and end of the selected output channel.
    The measurement is then run, and the time lag between
    a selected acquired signal and the output signal is computed
    from cross-correlation calculation.
    All the acquired signals are then re-synced from the time lag value.

    :param M: The measurement object
    :type M: measpy.measurement.Measurement
    :param out_chan: The selected output channel for synchronization. It is the index of the selected output signal in the list ``M.out_sig``
    :type out_chan: int
    :param in_chan: The selected input channel for synchronization. It is the index of the selected input signal in the list ``M.in_sig``
    :type in_chan: int
    :param added_time: Duration of silence added before and after the selected output signal
    :type added_time: float
    :return: Measured delay between i/o sync channels
    :rtype: float
    """
    M.sync_prepare(out_chan=out_chan,added_time=added_time)
    ni_run_measurement(M)
    d = M.sync_render(in_chan=in_chan,out_chan=out_chan,added_time=added_time)
    return d

def ni_get_devices():
    """
    Get the list of NI devices present in the system

    :returns: A list of devices object
    """
    system = nidaqmx.system.System.local()
    print(system.devices)
    return system.devices



# measpy/measurement.py
#
# ------------------------------------------------------
# Measuremnt class definition for data acquisition tasks
# ------------------------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

from .signal import Signal
from functools import partial

from ._tools import (csv_to_dict, 
                     convl, 
                     convl1,  
                     calc_dur_siglist,
                     h5file_write_from_queue)

import numpy as np
import matplotlib.pyplot as plt

from copy import copy

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from copy import copy,deepcopy
import csv
import os
import h5py

class Measurement:
    # ---------------------------
    def __init__(self, **params):
        # Check out_sig contents
        if 'out_sig' in params:
            non=params['out_sig']!=None
            sig=type(params['out_sig'])!=list
            if non&sig:
                raise Exception("out_sig must but be a list of measpy.signal.Signal or None")
            else:
                #print(list((type(s)==Signal for s in params['out_sig'])))
                if all((type(s)==Signal for s in params['out_sig'])):
                    # print('These are all signals')
                    if all(s.fs==params['out_sig'][0].fs for s in params['out_sig']):
                        # print('Same fs for all signals')
                        self.out_sig = params['out_sig']
                    else:
                        raise Exception("Signals in out_sig list have different sampling frequencies")                     
                else:
                    raise Exception("Some elements of out_sig list are not Signals")
        else:
            self.out_sig = None

        # Check in_sig contents
        if 'in_sig' in params:
            non=params['in_sig']!=None
            sig=type(params['in_sig'])!=list
            if non&sig:
                raise Exception("in_sig must but be a list of measpy.signal.Signal or None")
            else:
                if all((type(s)==Signal for s in params['in_sig'])):
                    # print('These are all signals')
                    if all(s.fs==params['in_sig'][0].fs for s in params['in_sig']):
                        # print('Same fs for all signals')
                        self.in_sig = params['in_sig']
                    else:
                        raise Exception("Signals in in_sig list have different sampling frequencies")                     
                else:
                    raise Exception("Some elements of in_sig list are not Signals")
        else:
            self.in_sig = None

        #Check sampling frequencies
        if type(self.out_sig)==type(None):
            if type(self.in_sig)==type(None):
                #raise Exception("This is a task with no input nor output ?")
                print("This is a task with no input nor output ?")
            else:
                # print("This is a task with no output.")
                if 'fs' in params:
                    if params['fs']!=self.in_sig[0].fs:
                        print('Selected sampling frequency '+str(params['fs'])+'Hz is different to that given in signals in in_sig list')
                        print('Sampling frequencies of all input signals are set to the selected value: '+str(params['fs'])+'Hz.')
                        self.fs = params['fs']
                        for s in self.in_sig:
                            s.fs = params['fs']
                    else:
                        self.fs = params['fs']
                else:
                    self.fs = self.in_sig[0].fs
                    print('Task frequency is: ', str(self.fs) )
        else:
            if 'fs' in params:
                if params['fs']!=self.out_sig[0].fs:
                    print('Selected sampling frequency '+str(params['fs'])+'Hz is different to that given in signals in out_sig list')
                    print("Task's sampling frequency is set to the selected value in outpu signals: "+str(self.out_sig[0].fs)+"Hz.")
            self.fs = self.out_sig[0].fs
            if type(self.in_sig)==type(None):
                print ("This is a task with no input.")
            else:
                if self.fs!=self.in_sig[0].fs:
                    print('Selected sampling frequency '+str(self.fs)+'Hz is different to that given in signals on in_sig list')
                    print('Sampling frequencies of all input signals are set to the selected value: '+str(self.fs)+'Hz.')
                    for s in self.in_sig:
                        s.fs = self.fs

        # Check list lengths
        if type(self.out_sig)!=type(None):
            if 'out_map' in params:
                if len(params['out_map'])!=len(self.out_sig):
                    raise Exception('Lengths of out_map and out_sig do not correspond.')
                self.out_map = params['out_map']
            else:
                self.out_map = list(range(1,len(self.out_sig)+1))
                print("out_map not given, it is set to default value of: "+str(self.out_map))
        if type(self.in_sig)!=type(None):
            if 'in_map' in params:
                if len(params['in_map'])!=len(self.in_sig):
                    raise Exception('Lengths of in_map and in_sig do not correspond.')
                self.in_map = params['in_map']
            else:
                self.in_map = list(range(1,len(self.in_sig)+1))
                print("in_map not given, it is set to default value of: "+str(self.in_map))
                print(self.in_map)
            if 'in_threshold' in params:
                if len(params['in_threshold'])!=len(self.in_sig):
                    raise Exception('Lengths of in_threshold and in_sig do not correspond.')
                self.in_threshold = params['in_threshold']

        self.in_device = params.setdefault("in_device",'')
        self.out_device = params.setdefault("out_device",'')

        # Check durations
        if 'dur' in params:
            if type(self.out_sig)==type(None):
                self.dur = params['dur']
            else:
                dursigs = calc_dur_siglist(self.out_sig)
                if params['dur']!=dursigs:
                    print('Selected duration is different thant duration of combined output signals.')
                    print('It is changed to match.')
                self.dur = dursigs
        else:
            if type(self.out_sig)==type(None):
                #raise Exception('No duration nor out_sig given. Impossible to determine task duration')
                print('No duration nor out_sig given. Impossible to determine task duration')
            else:
                self.dur = calc_dur_siglist(self.out_sig)
                print("Duration of the task set to: "+str(self.dur)+" s.")

        if 'device_type' not in params:
            self.device_type = ''
            print('No device_type given, it is set to empty string and will be updated when performing the task.')
        else:
            self.device_type = params['device_type']

        # Fix specific properties
        if self.device_type=='pico':
            self.in_range = params.setdefault("in_range",list('10V' for b in self.in_map))
            self.upsampling_factor = params.setdefault("upsampling_factor",1)
            self.in_coupling = params.setdefault("in_coupling",list('dc' for b in self.in_map))
            self.sig_gen = params.setdefault("sig_gen",False)
            if self.sig_gen != None:
                self.offset = params.setdefault("offset",0.0)
                self.wave = params.setdefault("wave",0)
                self.amp = params.setdefault("amp",1.0)
                self.freq_start = params.setdefault("freq_start",20)
                self.freq_stop = params.setdefault("freq_start",20_000)
                self.freq_change = params.setdefault("freq_change",10)
                self.freq_int = params.setdefault("freq_int",0.01)
                self.sweep_dir = params.setdefault("sweep_dir",0)
                self.sweep_number = params.setdefault("sweep_number",100)
        if self.device_type=='ni':
            self.in_range = params.setdefault("in_range",None)
            self.out_range = params.setdefault("out_range",None)
            self.in_iepe = params.setdefault("in_iepe",list(False for b in self.in_map))
        if type(self.out_sig)!=type(None):
            self.io_sync = params.setdefault('io_sync',0)
        elif 'io_sync' in params:
                print('No output signals given. io_sync param ignored')
        self.desc = params.setdefault('desc','No description')

    # -----------------
    def __repr__(self):
        out = "measpy.Measurement("
        out += "fs="+str(self.fs)
        out += ",\n dur="+str(self.dur)
        out += ",\n device_type='"+str(self.device_type)+"'"
        if self.in_sig!=None:
            out += ',\n in_device='+str(self.in_device)
            out += ',\n in_map='+str(self.in_map)
        try:
            out += ",\n date='"+self.date+"'"
            out += ",\n time='"+self.time+"'"
        except:
            pass
        if self.out_sig!=None:
            out += ",\n out_device='"+str(self.out_device)+"'"
            out += ',\n out_map='+str(self.out_map)
            out += ',\n out_sig=list of '+str(len(self.out_sig))+' measpy.signal.Signal'
            out += ",\n io_sync="+str(self.io_sync)
        if self.device_type=='pico':
            out += ",\n in_range="+str(self.in_range)
            out += ",\n upsampling_factor="+str(self.upsampling_factor)
            out += ",\n in_coupling="+str(self.in_coupling)
        if self.device_type=='ni':
            out += ",\n in_range="+str(self.in_range)
            out += ",\n out_range="+str(self.out_range)
            out += ",\n in_iepe="+str(self.in_iepe)            
        out += ',\n in_sig=list of '+str(len(self.in_sig))+' measpy.signal.Signal'
        out +=")"
        
        return out
    
    # ------------------------------
    def _to_dict(self,withsig=True):
        """ Converts a Measurement object to a dict

            :param withsig: Optionnally removes the data arrays, defaults to True
            :type withsig: bool
            
        """
        mesu = copy(self.__dict__)
        if not(withsig):
            del mesu['in_sig']
            del mesu['out_sig']
        return mesu

    # -----------------------
    def to_dir(self,dirname):
        """ Stores the parameters and signals in a directory

            :param dirname: Name of the directory, a (1), (2)... is added to the name if directory exists
            :type dirname: str
            :return: Actual name to the saved folder (if name conflit is detected)
            :rtype: str                            
        """
        if os.path.exists(dirname):
            i = 1
            while os.path.exists(dirname+'('+str(i)+')'):
                i+=1
            dirname = dirname+'('+str(i)+')'
        os.mkdir(dirname)
        self._params_to_csv(dirname+"/params.csv")
        if type(self.in_sig)!=type(None):
            for i,s in enumerate(self.in_sig):
                s.to_csvwav(dirname+"/in_sig_"+str(i))
        if type(self.out_sig)!=type(None):
            for i,s in enumerate(self.out_sig):
                s.to_csvwav(dirname+"/out_sig_"+str(i))
        self._write_readme(dirname+"/README")
        return dirname
    
    def to_hdf5(self, filename):
        """
        Save Measurement in hdf5 file

        :param filename: hdf5 file
        :type filename: str

        """
        mesu = self._to_dict()
        in_sig = mesu.pop("in_sig",None)
        out_sig = mesu.pop("out_sig",None)
        datatype = mesu.pop("datatype",None)
        multichannel = mesu.pop("multichannel",False)
        with h5py.File(filename, "x") as H5file:
            for name, value in mesu.items():
                if value is not None:
                    H5file.attrs[name] = value
            if type(in_sig)!=type(None):
                if multichannel:
                    raise NotImplementedError
                    in_sig.to_hdf5(H5file, "in_sigs", datatype)
                # elif (Nchannel := len(mesu["in_map"])) > 1:
                #     itemsize = Nchannel*np.dtype(datatype).itemsize
                #     #Chunck memory size should be between 10KiB and 1MiB, (bytes power of two 14 to 19 )
                #     power_two_chunck_size = 17
                #     chunksize = 2**(power_two_chunck_size-(itemsize-1).bit_length())
                #     dataset = H5file.create_dataset(
                #         "in_sigs", (0,Nchannel), maxshape=(None,Nchannel), dtype=datatype, chunks=(chunksize,Nchannel)
                #     )
                #     temp_dic = {}
                #     for index,s in zip(mesu["in_map"],in_sig):
                #         temp_dic.setdefault("channel",[]).append(index)
                #         for key, value in s.__dict__.items():
                #             if key not in ['_rawvalues',"h5save_data"]:
                #                 try:
                #                     Val = value.__str__()
                #                 except AttributeError:
                #                     Val = value
                #                 temp_dic.setdefault(key,[]).append(Val)
                #         for key, value in temp_dic.items():
                #             dataset.attrs[key] = value
                #     self.h5save_data = partial(
                #         h5file_write_from_queue,
                #         filename=H5file.filename,
                #         dataset_name="in_sigs",
                #         )
                else:
                    for index,s in zip(mesu["in_map"],in_sig):
                        s.to_hdf5(H5file, f"in_sig_{index}", datatype)
            if type(out_sig)!=type(None):
                if multichannel:
                    raise NotImplementedError
                    out_sig.to_hdf5(H5file, "out_sigs", datatype)
                else:
                    for index,s in zip(mesu["out_map"],out_sig):
                        s.to_hdf5(H5file, f"out_sig_{index}", datatype)

    # ------------------------
    @classmethod
    def from_dir(cls,dirname):
        """ Load a measurement object from a directory

            :param dirname: Name of the directory
            :type dirname: str                
        """

        task_dict = csv_to_dict(dirname+'/params.csv')
        self = cls._from_dict(task_dict)
        if 'in_map' in task_dict:
            self.in_sig = list(Signal.from_csvwav(dirname+'/in_sig_'+str(i)) for i in range(len(task_dict['in_map'])) )
        if 'out_map' in task_dict:
            self.out_sig = list(Signal.from_csvwav(dirname+'/out_sig_'+str(i)) for i in range(len(task_dict['out_map'])) )
        return self

    @classmethod
    def from_hdf5(cls,filename):
        with h5py.File(filename, "r") as H5file:
            task_dict = {}
            for key,val in H5file.attrs.items():
                task_dict[key] = val
            self = cls._from_dict(task_dict)
            if 'in_map' in task_dict:
                index = task_dict["in_map"]
                in_sig = [None]*len(index)
                try:
                    in_sig = [Signal.from_hdf5(H5file["in_sigs"],chan) for chan in index]
                except KeyError:
                    for i, ind in enumerate(index):
                        in_sig[i] = Signal.from_hdf5(H5file[f"in_sig_{ind}"])
            else:
                in_sig = None
            if 'out_map' in task_dict:
                index = task_dict["out_map"]
                out_sig = [None]*len(index)
                for i, ind in enumerate(index):
                    out_sig[i] = Signal.from_hdf5(H5file[f"out_sig_{ind}"])
            else:
                out_sig = None
            self.in_sig = in_sig
            self.out_sig = out_sig
        return self

    @classmethod
    def _from_dict(cls, task_dict):
        self=cls()
        self.fs=convl1(float,task_dict['fs'])
        self.dur=convl1(float,task_dict['dur'])
        try:
            self.date=convl1(str,task_dict['date'])
            self.time=convl1(str,task_dict['time'])
        except:
            pass
        self.device_type=convl1(str,task_dict['device_type'])

        if 'in_map' in task_dict:
            self.in_map = convl(int,task_dict['in_map'])
            self.in_device = convl1(str,task_dict['in_device'])

        else:
            self.in_sig = None
        if 'out_map' in task_dict:
            self.out_map = convl(int,task_dict['out_map'])
            self.out_device = convl1(str,task_dict['out_device'])
            self.io_sync = convl1(int,task_dict['io_sync'])
        else:
            self.out_sig = None

        # Picoscope specifics
        if self.device_type == 'pico':
            try:
                self.in_range = convl(str,task_dict['in_range'])
            except:
                pass
            try:
                self.in_coupling = convl(str,task_dict['in_coupling'])
            except:
                pass
            try:
                self.upsampling_factor = convl1(int,task_dict['upsampling_factor'])
            except:
                pass
            try:
                self.sig_gen = convl1(bool,task_dict['sig_gen'])
            except:
                pass
            try:
                self.offset = convl1(float,task_dict['offset'])
            except:
                pass
            try:
                self.wave = convl1(int,task_dict['wave'])
            except:
                pass
            try:
                self.freq_start = convl1(float,task_dict['freq_start'])
            except:
                pass
            try:
                self.freq_stop = convl1(float,task_dict['freq_stop'])
            except:
                pass
            try:
                self.freq_change = convl1(float,task_dict['freq_change'])
            except:
                pass
            try:
                self.freq_int = convl1(float,task_dict['freq_int'])
            except:
                pass
            try:
                self.sweep_dir = convl1(int,task_dict['sweep_dir'])
            except:
                pass
            try:
                self.sweep_number = convl1(float,task_dict['sweep_number'])
            except:
                pass

        # NI specifics
        if self.device_type == 'ni':
            try:
                self.in_range = convl1(float,task_dict['in_range'])
            except:
                pass
            try:
                self.out_range = convl1(float,task_dict['out_range'])
            except:
                pass
            try:
                self.in_iepe = convl(bool,task_dict['in_iepe'])
            except:
                pass

        return self

    # --------------------------------
    def _params_to_csv(self,filename):
        """ Writes all the Measurement object parameters to a csv file """
        dd = self._to_dict(withsig=False)
        with open(filename, 'w', newline='') as file:
            writer = csv.writer(file)
            for key in dd:
                if type(dd[key])==list:
                    writer.writerow([key]+dd[key])
                else:
                    writer.writerow([key,str(dd[key])])

    # -------------------------------
    def _write_readme(self,filename):
        with open(filename, 'w') as f:
            f.write('Created with measpy version '+__version__)
            f.write('\n')
            f.write('https://github.com/odoare/measpy')

    def sync_prepare(self,out_chan=0,added_time=1):
        """
        Prepare measurement for synchronization

        :param out_chan: The selected output channel for synchronization. It is the index of the selected output signal in the list ``M.out_sig``
        :type out_chan: int
        :param in_chan: The selected input channel for synchronization. It is the index of the selected input signal in the list ``M.in_sig``
        :type in_chan: int
        :param added_time: Duration of silence added before and after the selected output signal
        :type added_time: float
        """        
        osig = self.out_sig[out_chan].add_silence((added_time,added_time)).delay(-added_time)
        self.dur = self.dur+2*added_time
        self.out_sig[out_chan]=osig

    def sync_render(self,out_chan=0,in_chan=0,added_time=1):
        d = self.in_sig[in_chan].timelag(self.out_sig[out_chan])
        dt = 1/self.fs
        # print("delay: "+str(d)+"s")
        self.dur = self.dur-2*added_time
        self.out_sig[out_chan] = self.out_sig[out_chan].cut(dur=(added_time,added_time+self.dur)).delay(added_time)
        for i,s in enumerate(self.in_sig):
            self.in_sig[i] = s.cut(dur=(added_time+d+dt,added_time+d+dt+self.dur))
            self.in_sig[i].t0 = self.out_sig[out_chan].t0
        return d

# def peak_sync_prepare(M,out_chan=0,in_chan=0):
#     def inner(*args,**kwargs):
#         osig = M.out_sig[out_chan]
#         osigpeak = osig.similar(
#             values=np.stack((picv(long=2*M.fs),osig.values)),
#             t0=osig.t0-2)
#         M1 = copy(M)
#         M1.out_sig[out_chan]=osigpeak
#         func(M1)
#         isig = M1.in_sig[in_chan]
#         posmax = int( np.argmax(isig.values[int(0.25*M.fs*2):int(0.75*M.fs*2)]) + 0.75*M.fs*2 )
#         print(posmax)
#         for i,s in enumerate(M1.in_sig):
#             s.values = s.values[posmax:posmax+M.fs*M.dur]
#         M=M1
#         del(M1)
#     return inner

    # def peak_sync_prepare(self,out_chan=0):
    #     osig = self.out_sig[out_chan]
    #     osigpeak = osig.similar(
    #         values=np.stack((picv(long=2*M.fs),osig.values)),
    #         t0=osig.t0-2)
    #     del(osig)
    #     M1 = deepcopy(self)
    #     M1.out_sig[out_chan]=osigpeak
    #     return M1

    # def peak_sync_render(self,in_chan=0):
    #     isig = self.in_sig[in_chan]
    #     posmax = int( np.argmax(isig.values[int(0.25*M.fs*2):int(0.75*M.fs*2)]) + 0.75*M.fs*2 )
    #     print(posmax)
    #     for i,s in enumerate(M1.in_sig):
    #         s.values = s.values[posmax:posmax+M.fs*M.dur]
    #     return M1
    #     del(M1)




# __init__.py

from .signal import (Signal, 
                            Spectral,
                            Weighting,
                            PREF,
                            DBUREF,
                            DBVREF,
                            WDBA,
                            WDBC,
                            WDBM)
from .measurement import Measurement

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# measpy/pico.py
#
# ---------------------------------------
# Data acquisition with Picoscope devices
# ---------------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy


import numpy as np
from inspect import getfullargspec

from datetime import datetime
from scipy.signal import decimate

import ctypes
from picosdk.ps4000 import ps4000
from picosdk.functions import adc2mV, assert_pico_ok

from threading import Thread
from queue import Queue, Empty
import time
from scipy.signal import find_peaks
from functools import partial

from picosdk.ps2000 import ps2000
from picosdk.functions import assert_pico2000_ok
from picosdk.ctypes_wrapper import C_CALLBACK_FUNCTION_FACTORY
from picosdk.PicoDeviceEnums import picoEnum

import matplotlib.pyplot as plt
from matplotlib.widgets import Button

# plt.style.use('seaborn-v0_8')

maxtimeout = 10

PS2000_channel = {"A": 1, "B": 2, 1: "A", 2: "B"}

class inline_plotting:
    # """
    # Plot data from a Queue.queue
    # init create the plot
    # update_plot update the plot buffer from new data in dataqueues and update the plot if enough data has been received
    # end_plot call update plot until dataqueues contain None
    # ploting_duration estimate time of update_plot execution to check for buffer overflow risk
    # """

    def __init__(self, fs, timeout, updatetime=0.1, plotbuffersize=2000):
        self.timesincelastupdate = 0.0
        self.plotbuffersize = plotbuffersize
        self.updatetime = updatetime
        self.fs = fs
        self.timeinterval = 1/fs
        self.timeout = min(timeout,0.1 * updatetime)
        self.stop = False

        def fstop(event):
            global stop
            self.stop = True

        def tamp_plus(event):
            self.ax[0].set_ylim(np.array(self.ax[0].get_ylim()) / 2)
            plt.pause(0.0001)

        def tamp_moins(event):
            self.ax[0].set_ylim(np.array(self.ax[0].get_ylim()) * 2)
            plt.pause(0.0001)

        def famp_plus(event):
            self.ax[1].set_ylim(np.array(self.ax[1].get_ylim()) / 2)
            plt.pause(0.0001)

        def famp_moins(event):
            self.ax[1].set_ylim(np.array(self.ax[1].get_ylim()) * 2)
            plt.pause(0.0001)

        def on_close(event):
            global stop
            stop = True

        self.x = 1.0 * np.arange(-self.plotbuffersize, 0) * self.timeinterval
        self.plotbuffer = np.zeros_like(self.x)
        self.fig, self.ax = plt.subplots(1, 2, figsize=(15, 5))
        self.fig.subplots_adjust(bottom=0.2)
        self.ax[0].set_xlabel("Temps [s]", fontsize=15)
        self.ax[0].set_ylabel("Tension [V]", fontsize=15)
        self.ax[1].set_xlabel("Fréquence [Hz]", fontsize=15)
        self.ax[1].set_ylabel("Tension [V]", fontsize=15)
        self.ax[1].set_xlim([0.01, self.fs / 2])
        self.ax[1].set_ylim([0, 10])
        (self.linet,) = self.ax[0].plot(self.x, self.plotbuffer)
        (self.linef,) = self.ax[1].plot(
            np.fft.fftfreq(n=self.plotbuffersize, d= self.timeinterval),
            np.fft.fft(self.plotbuffer, norm="ortho"),
        )
        axs = self.fig.add_axes([0.4, 0.01, 0.2, 0.075])
        self.bstop = Button(axs, "Stop")
        self.bstop.on_clicked(fstop)
        atxp = self.fig.add_axes([0.02, 0.6, 0.04, 0.05])
        self.btplus = Button(atxp, "+")
        self.btplus.on_clicked(tamp_plus)
        atxm = self.fig.add_axes([0.02, 0.4, 0.04, 0.05])
        self.btmoins = Button(atxm, "-")
        self.btmoins.on_clicked(tamp_moins)
        afxp = self.fig.add_axes([0.92, 0.6, 0.04, 0.05])
        self.bfplus = Button(afxp, "+")
        self.bfplus.on_clicked(famp_plus)
        afxm = self.fig.add_axes([0.92, 0.4, 0.04, 0.05])
        self.bfmoins = Button(afxm, "-")
        self.bfmoins.on_clicked(famp_moins)
        self.fig.canvas.mpl_connect("close_event", on_close)
        plt.pause(0.0001)

    def _plotting_buffer(self):
        self.x += self.timesincelastupdate
        self.linet.set_ydata(self.plotbuffer)
        self.linet.set_xdata(self.x)
        self.ax[0].relim()
        self.ax[0].autoscale_view()
        self.linef.set_ydata(
            np.abs(
                np.fft.fft(
                    self.plotbuffer * np.hanning(self.plotbuffersize),
                    norm="ortho",
                )
            )
        )
        plt.pause(0.01)
        self.timesincelastupdate = 0.0

    def _update_buffer(self, item):
        n_values = len(item)
        item = np.asarray(item)*0.001
        self.timesincelastupdate += n_values * self.timeinterval
        self.plotbuffer = np.roll(self.plotbuffer, int(-n_values), axis=0)
        self.plotbuffer[-n_values:] = item[-self.plotbuffersize :]

    def update_plot(self, dataqueue, updatetime=None):
        updatetime = self.updatetime if updatetime is None else updatetime
        try:
            if (item := dataqueue.get(timeout=self.timeout)) is not None:
                self._update_buffer(item)
                if self.timesincelastupdate > updatetime:
                    self._plotting_buffer()
        except Empty:
            pass

    def end_plot(self, dataqueue):
        try:
            while (item := dataqueue.get(timeout=maxtimeout)) is not None:
                self._update_buffer(item)
                if self.timesincelastupdate > self.updatetime:
                    self._plotting_buffer()
            self._plotting_buffer()
        except Empty:
            pass
        print("End of datastream")

    @property
    def ploting_duration(self):
        try:
            return self._ploting_duration
        except AttributeError:
            queuetest = Queue()
            queuetest.put(np.random.normal(0, 3, self.plotbuffersize))
            queuetest.put(None)
            start = time.time()
            self.update_plot(queuetest, updatetime=0)
            self._ploting_duration = time.time() - start
            print(f"Plotting duration = {self._ploting_duration}")
            self.x = 1.0 * np.arange(-self.plotbuffersize, 0) * self.timeinterval
            self.timesincelastupdate = 0
            self._plotting_buffer()
            return self._ploting_duration


def findindex(l, e):
    try:
        a = l.index(e)
    except:
        a = None
    return a


def detect_rising_pulses_grad_ind(values, ind0, previous_data_points, **kwargs):
    # """
    # Detect rising pulse with signal.find_peaks on gradient of the signal

    # Parameters
    # ----------
    # values : List
    #     Data.
    # ind0 : int
    #     indice of the first data point.
    # previous_data_points : number
    #     Value of the last data point (indice = ind0-1).
    #     useful to not lose peak if it is at ind0 (work not very well)
    # range_ : int
    #     picoscope range.

    # Returns
    # -------
    # List
    #     List of indices where a rising pulse is detected.

    # """
    try:
        N = len(previous_data_points)
        previous_data_points.extend(values)
        grad = np.diff(previous_data_points)
        lefts, _ = find_peaks(grad, height=np.mean(grad) + np.std(grad))
        # rights, _ = find_peaks(-grad, height=2000)
        return lefts + ind0 - N + 1
    except TypeError:
        grad = np.diff(values)
        lefts, _ = find_peaks(grad, height=np.mean(grad) + np.std(grad))
        # rights, _ = find_peaks(-grad, height=2000)
        return lefts + ind0 + 1
    except AttributeError:
        grad = np.diff([previous_data_points] + values)
        lefts, _ = find_peaks(grad, height=np.mean(grad) + np.std(grad))
        # rights, _ = find_peaks(-grad, height=2000)
        return lefts + ind0


def detect_rising_pulses_threshold_ind(
    values, ind0, previous_data_point, threshold, **kwargs
):
    # """
    # Detect rising pulse using a threshold
    # Parameters
    # ----------
    # values : List
    #     Data.
    # ind0 : int
    #     indice of the first data point.
    # previous_data_points : number
    #     Value of the last data point (indice = ind0-1).
    #     useful to not lose peak if it is at ind0
    # range_ : int
    #     picoscope range.
    # threshold : int
    #     threshold (in adc values).

    # Returns
    # -------
    # List
    #     List of indices where a rising pulse is detected.

    # """
    try:
        V = np.asarray([previous_data_point] + values)
        rising = np.flatnonzero((V[:-1] <= threshold) & (V[1:] > threshold))
    except TypeError:
        V = np.asarray(values)
        rising = np.flatnonzero((V[:-1] <= threshold) & (V[1:] > threshold)) + 1
    return rising + ind0


def rising_pulse_to_raw(M, channel, values):
    for i in range(len(M.in_map)):
        if M.in_map[i] == channel:
            M.in_sig[i].fs = None
            M.in_sig[i].raw = np.double([v / M.fs for v in values])
            M.in_sig[i].desc = "Timming of rising edge pulse"
            M.in_sig[i].unit = "s"


def adc_to_mv(values, range_, bitness=16, **kwargs):
    v_ranges = [10, 20, 50, 100, 200, 500, 1_000, 2_000, 5_000, 10_000, 20_000, 50_000, 100_000]
    return [(x * v_ranges[range_]) / (2 ** (bitness - 1) - 1) for x in values]


def mv_to_adc(values, range_, bitness=16):
    v_ranges = [10, 20, 50, 100, 200, 500, 1_000, 2_000, 5_000, 10_000, 20_000, 50_000, 100_000]
    return [
        int((x * (2 ** (bitness - 1) - 1)) / v_ranges[range_]) for x in values
    ]


def mv_to_raw(M, channel, values):
    for i in range(len(M.in_map)):
        if M.in_map[i] == channel:
            M.in_sig[i].fs = M.fs
            if M.upsampling_factor > 1:
                M.in_sig[i].raw = decimate(
                    np.double(values) * 0.001, M.upsampling_factor, ftype="fir"
                )[0 : int(round(M.dur * M.fs))]
            else:
                M.in_sig[i].raw = (np.double(values) * 0.001)[
                    0 : int(round(M.dur * M.fs))
                ]


def ps2000_run_measurement(M):
    return _ps2000_run_measurement_threaded(M, adc_to_mv, mv_to_raw, None)


def ps2000_plot(M, plotbuffersize=2000, updatetime=0.1, chan_to_plot="A"):
    plotting = partial(inline_plotting, plotbuffersize=plotbuffersize, updatetime=updatetime)
    return _ps2000_run_measurement_threaded(
        M,
        adc_to_mv,
        mv_to_raw,
        plotting,
        chan_to_plot=chan_to_plot,
    )


def ps2000_pulse_detection(M):
    # """
    # To use threshold detection (more efficient), this function needs M to contain the property
    #  in_threshold : a list of threshold for pulse height (in Volt) for each channel
    # """
    if hasattr(M, "in_threshold") and M.in_threshold is not None:
        detect_rising_pulses = detect_rising_pulses_threshold_ind
        info = ", ".join([f"Threshold = {th} V for chan {PS2000_channel[chan]}" for chan,th in zip(M.in_map,M.in_threshold)])
        print(
            f"Detect pulse with threshold method ({info})"
        )
    else:
        detect_rising_pulses = detect_rising_pulses_grad_ind
        print("Detect pulse with gradient method")
    return _ps2000_run_measurement_threaded(
        M,
        detect_rising_pulses,
        rising_pulse_to_raw,
        None,
        min_chunksize_processed=10000,
    )

def ps2000_hdf5(M,filename):
    return _ps2000_run_measurement_threaded(M,None,None,None,filename=filename)

def _ps2000_run_measurement_threaded(
    M,
    pre_process,
    save,
    plotting,
    chan_to_plot=None,
    min_chunksize_processed=0,
    filename = None
):
    # """
    # This function needs M to contain the following properties:
    #     - in_range : a list of strings specifying the voltage range.
    #         Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
    #         "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
    #     - upsampling_factor : upsampling factor
    #     - in_coupling : Coupling configuration of the channels.
    #         Can be "ac" or "dc"
    # """

    savehdf5 = False
    if filename is not None:
        print(f"Measurment will be save in {filename}, plotting disabled, preprocess and save method ignored")
        savehdf5 = True

    multichannel = False
    max_samples = 100_000
    overview_buffer_size = 20_000

    CALLBACK = C_CALLBACK_FUNCTION_FACTORY(
        None,
        ctypes.POINTER(ctypes.POINTER(ctypes.c_int16)),
        ctypes.c_int16,
        ctypes.c_uint32,
        ctypes.c_int16,
        ctypes.c_int16,
        ctypes.c_uint32,
    )

    if M.device_type != "pico":
        raise Exception("Error: device_type must be 'pico'.")

    if type(M.out_sig) != type(None):
        print("Warning: out_sig property ignored with picoscopes")

    # Effective sampling frequency
    # If upsampling_factor is > 1, the actual data acquisition is
    # performed at fs*upsampling_factor, and then decimated to
    # the desired frequency

    effective_fs = M.fs * M.upsampling_factor
    duree_ns = M.dur * 1e9

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments

    si = round(1e9 / effective_fs)
    if effective_fs != (1e9 / si):
        effective_fs = 1e9 / si
        print(
            "Warning : Effective sampling frequency fs changed to nearest possible value of "
            + str(effective_fs)
            + " Hz"
        )

    # print("Effective sampling frequency: " + str(effective_fs))
    sampleInterval = int(si)

    if M.fs != effective_fs / M.upsampling_factor:
        M.fs = effective_fs / M.upsampling_factor
        for s in M.in_sig:
            s.fs = M.fs
        print(
            "Warning : Signal sampling frequency fs changed to nearest possible value of "
            + str(M.fs)
            + " Hz"
        )

    max_loop_time = overview_buffer_size / M.fs
    if plotting is not None:
        if (
            chan_to_plot is not None
            and findindex(M.in_map, PS2000_channel.get(chan_to_plot)) is None
        ):
            print(
                f"The channel to plot {chan_to_plot} is not enabled in measurement setup, plotting canceled"
            )
            plotting = None
        else:
            if callable(plotting):
                plot = plotting(
                    M.fs,
                    timeout=0.1 * max_loop_time,
                )
            else:
                plot = inline_plotting(
                    M.fs,
                    timeout=0.1 * max_loop_time,
                )
            if plot.ploting_duration > 0.85 * max_loop_time:
                print(
                    "Update plot function is too long to execute : "
                    f"{plot.ploting_duration} s, there is a high risk to"
                    f" overflow the buffer (the buffer is full in {max_loop_time} s) "
                )
                plt.close(plot.fig)
                plotting = None
                if input("Cancel measurment? y/n : ") == "y":
                    return

    ## Setup channels
    def setup_channel(chan_index):
        if (ind := findindex(M.in_map, chan_index)) != None:
            enabled = True
            pico_range = ps2000.PS2000_VOLTAGE_RANGE["PS2000_" + M.in_range[ind]]
            if M.in_coupling[ind].capitalize() == "Dc":
                coupling = "PICO_DC"
            elif M.in_coupling[ind].capitalize() == "Ac":
                coupling = "PICO_AC"
            else:
                print(f"Input {PS2000_channel[chan_index]} coupling not recognized, set to 'dc'")
                M.in_coupling[ind] = "dc"
                coupling = "PICO_DC"
            print(
                f"Channel {PS2000_channel[chan_index]}: enabled with range "
                + "PS2000_"
                + M.in_range[ind]
                + " ("
                + str(pico_range)
                + ")"
            )
        else:
            enabled = False
            pico_range = ps2000.PS2000_VOLTAGE_RANGE["PS2000_10V"]
            coupling = "PICO_DC"
            print(f"Channel {PS2000_channel[chan_index]}: disabled")
        return enabled, coupling, pico_range

    enabledA, couplingA, rangeA = setup_channel(1)
    enabledB, couplingB, rangeB = setup_channel(2)

    def setup_save_hdf5(chan_index):
        # """
        # Setup to save directly on hdf5 file

        # Parameters
        # ----------
        # chan_index : int
        #     Index of he channel (1or 2).

        # Returns
        # -------
        # queue : queue.Queue
        #     Queue where data are sent by get_overview_buffers.
        # Process : method
        #     Preprocessing method to be sent in thread.

        # """
        queue = Queue()
        if isinstance(chan_index,list):
            method = M.h5save_data
            Process = Thread(target=method,args=(queue,))
        elif (ind := findindex(M.in_map, chan_index)) != None:
            Sig = M.in_sig[ind]
            method = Sig.h5save_data
            Process = Thread(target=method,args=(queue,))
        return queue, Process

    def setup_preprocess(chan_index, pico_range, chan_plot=False):
        # """
        # Define the method for preprocessing data in thread based on the method in arguments

        # Parameters
        # ----------
        # chan_index : int
        #     Index of the channel (1 or 2).
        # pico_range : int
        #     Picoscope voltage range (mapped from PS2000_VOLTAGE_RANGE).
        # chan_plot : bool, optional
        #     If True the data are send to another queue for plotting. The default is False.

        # Returns
        # -------
        # queue : queue.Queue
        #     Queue where data are sent by get_overview_buffers.
        # queue_plot : queue.Queue
        #     Queue for plotting.
        # Process : method
        #     Preprocessing method to be sent in thread.
        # result : list
        #     Data after preprocessing.

        # """
        if (ind := findindex(M.in_map, chan_index)) != None:
            result = []
            queue = Queue()

            if chan_plot:
                queue_plot = Queue()
            else:
                queue_plot = None

            if (Vthreshold := getattr(M, "in_threshold", None)) is not None:
                adc_threshold = mv_to_adc([Vthreshold[ind] * 0.001], pico_range)[0]
            else:
                adc_threshold = None

            # list pre_process argument, to choose wich method to use
            method_args = getfullargspec(pre_process).args

            #range_ and threshold, optianla argument that do not influence the func method
            kwargs = {"range_": pico_range, "threshold": adc_threshold}
            partial_kwargs = {
                a: b for a, b in kwargs.items() if a in method_args
            }
            if partial_kwargs:
                method = partial(pre_process, **partial_kwargs)
            else:
                method = pre_process

            if "ind0" and "previous_data_point" in method_args:
            #pre_process need the index of the start of each chunk and 
            #the value of the last datapoint (detect_rising_pulses_threshold_ind)
                def func(queue, result):
                    ind0 = 0
                    chunk_data = []
                    previous_data_point = None
                    while (item := queue.get(timeout=maxtimeout)) is not None:
                        chunk_data.extend(item)
                        if (N := len(chunk_data)) > min_chunksize_processed:
                            values = method(chunk_data, ind0, previous_data_point)
                            previous_data_point = chunk_data[-1]
                            result.extend(values)
                            ind0 += N
                            chunk_data = []
                    if len(chunk_data) > 0:
                        values = method(chunk_data, ind0, previous_data_point)
                        previous_data_point = chunk_data[-1]
                        result.extend(values)

            elif not ("ind0" in method_args or "previous_data_point" in method_args):
                if min_chunksize_processed > 0:
                # if preprocessing has to be done in large chuncks
                    if queue_plot is not None:

                        def func(queue, result):
                            chunk_data = []
                            while (item := queue.get(timeout=maxtimeout)) is not None:
                                chunk_data.extend(item)
                                if len(chunk_data) > min_chunksize_processed:
                                    values = method(chunk_data)
                                    result.extend(values)
                                    queue_plot.put(values)
                                    chunk_data = []
                            if len(chunk_data) > 0:
                                values = method(chunk_data)
                                result.extend(values)
                                queue_plot.put(values)
                            queue_plot.put(None)

                    else:
                        def func(queue, result):
                            chunk_data = []
                            while (item := queue.get(timeout=maxtimeout)) is not None:
                                chunk_data.extend(item)
                                if len(chunk_data) > min_chunksize_processed:
                                    values = method(chunk_data)
                                    result.extend(values)
                                    chunk_data = []
                            if len(chunk_data) > 0:
                                values = method(chunk_data)
                                result.extend(values)
                else:
                    if queue_plot is not None:
                        def func(queue, result):
                            while (item := queue.get(timeout=maxtimeout)) is not None:
                                values = method(item)
                                result.extend(values)
                                queue_plot.put(values)
                            queue_plot.put(None)

                    else:
                        def func(queue, result):
                            while (item := queue.get(timeout=maxtimeout)) is not None:
                                values = method(item)
                                result.extend(values)

            else:
                raise NotImplementedError(f"Pre_process arguments ({method_args}) is not supported")

            Process = Thread(target=func,args=(queue,result,))
            return queue, queue_plot, Process, result


    if savehdf5:
        #update dbfs before saving parameters
        for sig, prange in zip(M.in_sig,M.in_range):
            pico_range = ps2000.PS2000_VOLTAGE_RANGE["PS2000_" + prange]
            sig.dbfs = adc_to_mv([1],pico_range)[0] * 0.001

        #Picoscope2000 return 16bits signed integer
        M.datatype = "i2"
        M.to_hdf5(filename)
        if not multichannel:
            if enabledA:
                if enabledB:
                    queueAB, ProcessAB = setup_save_hdf5([1,2])
                    multichannel=True
                else:
                    queueA, ProcessA = setup_save_hdf5(1)
            elif enabledB:
                queueB, ProcessB = setup_save_hdf5(2)
    elif not multichannel:
        if enabledA:
            if chan_to_plot == "A":
                queueA, queue_plot, ProcessA, retA = setup_preprocess(
                    1, rangeA, True
                )
            else:
                queueA, _, ProcessA, retA = setup_preprocess(1, rangeA, False)

        if enabledB:
            if chan_to_plot == "B":
                queueB, queue_plot, ProcessB, retB = setup_preprocess(
                    2, rangeB, True
                )
            else:
                queueB, _, ProcessB, retB = setup_preprocess(2, rangeB, False)
    else:
        raise NotImplementedError #Need change to allow preprocess with multichannel
        # queueAB, queue_plot, ProcessAB, retAB = setup_preprocessAB(
        #     [1,2], [rangeA,rangeB], chan
        # )

    ##get_overview_buffers_factory
    if enabledA:
        if enabledB:
            if multichannel:
                def get_overview_buffers(
                    buffers, _overflow, _triggered_at, _triggered, _auto_stop, n_values
                ):
                    queueAB.put([buffers[0][0:n_values],buffers[2][0:n_values]])
            else:

                def get_overview_buffers(
                    buffers,
                    _overflow,
                    _triggered_at,
                    _triggered,
                    _auto_stop,
                    n_values,
                ):
                    queueA.put(buffers[0][0:n_values])
                    queueB.put(buffers[2][0:n_values])

        else:

            def get_overview_buffers(
                buffers,
                _overflow,
                _triggered_at,
                _triggered,
                _auto_stop,
                n_values,
            ):
                queueA.put(buffers[0][0:n_values])

    elif enabledB:

        def get_overview_buffers(
            buffers, _overflow, _triggered_at, _triggered, _auto_stop, n_values
        ):
            queueB.put(buffers[2][0:n_values])

    callback = CALLBACK(get_overview_buffers)

    with ps2000.open_unit() as device:
        print("Device info: {}".format(device.info))

        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL["PICO_CHANNEL_A"],
            enabledA,
            picoEnum.PICO_COUPLING[couplingA],
            rangeA,
        )
        assert_pico2000_ok(res)
        res = ps2000.ps2000_set_channel(
            device.handle,
            picoEnum.PICO_CHANNEL["PICO_CHANNEL_B"],
            enabledB,
            picoEnum.PICO_COUPLING[couplingB],
            rangeB,
        )
        assert_pico2000_ok(res)

        ########## Signal generator ##########
        if M.sig_gen:
            res = ps2000.ps2000_set_sig_gen_built_in(
                device.handle,
                int(M.offset * 1e6),  # offset voltage in uV
                int(M.amp / 2 * 1e6),  # peak-to-peak votage in uV
                M.wave,  # type of waveform (0 = sine wave)
                M.freq_start,  # start frequency in Hz
                M.freq_stop,  # stop frequency in Hz
                M.freq_change,  # frequency change per interval in Hz
                M.freq_int,  # interval of frequency change in seconds
                M.sweep_dir,  # sweep direction (0 = up)
                M.sweep_number,  # number of times to sweep
            )
            assert_pico2000_ok(res)
        #######################################

        res = ps2000.ps2000_run_streaming_ns(
            device.handle,
            sampleInterval,
            2,
            max_samples,
            False,
            1,
            overview_buffer_size,
        )
        assert_pico2000_ok(res)

        now = datetime.now()
        M.date = now.strftime("%Y-%m-%d")
        M.time = now.strftime("%H:%M:%S")

        ##start preprocess threads
        if multichannel:
            ProcessAB.start()
            if not savehdf5:
                ProcessA.start()
                ProcessB.start()
        else:
            if enabledA:
                ProcessA.start()

            if enabledB:
                ProcessB.start()

        print("Start")
        start_time = time.time_ns()

        # loop until wanted duration + maximum time of one loop without buffer overrun
        if plotting is not None:
            margin = max(min(int(max_loop_time*1e9),duree_ns * 0.1), plot.ploting_duration*(1.1))
            while time.time_ns() - start_time < duree_ns + margin and not plot.stop:
                ps2000.ps2000_get_streaming_last_values(device.handle, callback)
                plot.update_plot(queue_plot)

        else:
            margin = min(int(max_loop_time*1e9), duree_ns * 0.1)
            while time.time_ns() - start_time < duree_ns  + margin:
                ps2000.ps2000_get_streaming_last_values(device.handle, callback)

        print("Measurment done")
        print("Waiting for data processing...")
        overrun = False
        ps2000.ps2000_overview_buffer_status(device.handle, overrun)
        if overrun:
            print("Buffer have overrun")

        ##Put end flag in queues
        if multichannel:
            queueAB.put(None)
        else:
            if enabledA:
                queueA.put(None)
            if enabledB:
                queueB.put(None)

        ##Wait for all thread to finish and save data
        if multichannel:
            ProcessAB.join()
            if not savehdf5:
                ProcessA.join()
                save(M, 1, retA)
                ProcessB.join()
                save(M, 2, retB)
        else:
            if enabledA:
                ProcessA.join()
                if not savehdf5:
                    save(M, 1, retA)

            if enabledB:
                ProcessB.join()
                if not savehdf5:
                    save(M, 2, retB)

        print("Preprocess data done")

        ## Continue plotting until last data
        if plotting is not None:
            plot.end_plot(queue_plot)


def ps4000_plot(M,plotbuffersize=2000,updatetime=0.1):
    # """
    # This function needs M to contain the following properties:
    #     - in_range : a list of strings specifying the voltage range.
    #         Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
    #         "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
    #     - upsampling_factor : upsampling factor
    #     - in_coupling : Coupling configuration of the channels.
    #         Can be "ac" or "dc"
    # """

    import time


    # Plotting part

    global plotbuffer, ax, x, line, stop, timesincelastupdate, updtime

    updtime = updatetime

    timesincelastupdate = 0.0

    stop = False

    def fstop(event):
        global stop
        stop = True

    def tamp_plus(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())/2)
    def tamp_moins(event):
        ax[0].set_ylim(np.array(ax[0].get_ylim())*2)
    def famp_plus(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())/2)
    def famp_moins(event):
        ax[1].set_ylim(np.array(ax[1].get_ylim())*2)
    def on_close(event):
        global stop
        stop = True

    x = np.arange(plotbuffersize)
    plotbuffer = np.sin(x)
    fig, ax = plt.subplots(1,2,figsize=(15,5))
    fig.subplots_adjust(bottom=0.2)
    ax[0].set_xlabel('Temps [s]',fontsize=15)
    ax[0].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlabel('Fréquence [Hz]',fontsize=15)
    ax[1].set_ylabel('Tension [V]',fontsize=15)
    ax[1].set_xlim([0.01,M.fs/2])
    ax[1].set_ylim([0,10])
    linet, = ax[0].plot(x/M.fs,plotbuffer)
    linef, = ax[1].plot(np.fft.fftfreq(n=plotbuffersize,d=1/M.fs),np.fft.fft(plotbuffer,norm='ortho'))
    axs = fig.add_axes([0.4, 0.01, 0.2, 0.075])
    bstop = Button(axs, 'Stop')
    bstop.on_clicked(fstop)
    atxp = fig.add_axes([0.02, 0.6, 0.04, 0.05])
    btplus = Button(atxp, '+')
    btplus.on_clicked(tamp_plus)
    atxm = fig.add_axes([0.02, 0.4, 0.04, 0.05])
    btmoins = Button(atxm, '-')
    btmoins.on_clicked(tamp_moins)
    afxp = fig.add_axes([0.92, 0.6, 0.04, 0.05])
    bfplus = Button(afxp, '+')
    bfplus.on_clicked(famp_plus)
    afxm = fig.add_axes([0.92, 0.4, 0.04, 0.05])
    bfmoins = Button(afxm, '-')
    bfmoins.on_clicked(famp_moins)
    fig.canvas.mpl_connect('close_event', on_close)
    plt.pause(0.0001)

    # End of plotting part

    # Find maximum ADC count value
    # handle = chandle
    # pointer to value = ctypes.byref(maxADC)
    maxADC = ctypes.c_int16(32767)

    global nextSample, autoStopOuter, wasCalledBack

    # Buffer size fixed to 20k samples
    sizeOfOneBuffer = 20_000

    # Effective sampling frequency
    # If upsampling is > 1, the actual data acquisition is
    # performed at fs*upsampling, and then decimated to
    # the desired frequency
    effective_fs = M.fs * M.upsampling_factor

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments
    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    numdesiredsamples = int(round(effective_fs*M.dur))
    numBuffersToCapture = int(np.ceil(numdesiredsamples/sizeOfOneBuffer))
    sampleInterval = ctypes.c_int32(si)
    totalSamples = sizeOfOneBuffer * numBuffersToCapture
    chandle = ctypes.c_int16()
    status = {}

    # Open PicoScope 4000 Series device
    # Returns handle to chandle for use in future API functions
    status["openunit"] = ps4000.ps4000OpenUnit(ctypes.byref(chandle))
    assert_pico_ok(status["openunit"])

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indA]]
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 1
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 0
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 1
            M.in_coupling[indA]='dc'
        # Create buffers ready for assigning pointers for data collection
        bufferAMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteA = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel A: enabled with range '+'PS4000_'+M.in_range[indA]+' ('+str(rangeA)+')')
    else:
        enabledA = False
        rangeA = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel A: disabled')

    # Set up channel A
    channel_range = ps4000.PS4000_RANGE['PS4000_50MV']
    status["setChA"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                            int(enabledA),
                                            couplingA,
                                            rangeA)
    assert_pico_ok(status["setChA"])

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indB]]
        # Create buffers ready for assigning pointers for data collection
        bufferBMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteB = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel B: enabled with range '+'PS4000_'+M.in_range[indB]+' ('+str(rangeB)+')')
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 1
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 0
        else:
            print("Input B coupling not recognized, set to 'dc'")
            couplingB = 1
            M.in_coupling[indB]='dc'
    else:
        enabledB = False
        couplingB = 1
        rangeB = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel B: disabled')

    # Set up channel B
    status["setChB"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                            int(enabledB),
                                            couplingB,
                                            rangeB)
    assert_pico_ok(status["setChB"])

    # Set data buffer location for data collection from channel A
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_A = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledA:
        status["setDataBuffersA"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                                        bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersA"])

    # Set data buffer location for data collection from channel B
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_B = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledB:
        status["setDataBuffersB"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                                        bufferBMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersB"])

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Begin streaming mode:
    sampleUnits = ps4000.PS4000_TIME_UNITS['PS4000_NS']
    # We are not triggering:
    maxPreTriggerSamples = 0
    autoStopOn = 1
    # No downsampling:
    downsampleRatio = 1
    status["runStreaming"] = ps4000.ps4000RunStreaming(chandle,
                                                    ctypes.byref(sampleInterval),
                                                    sampleUnits,
                                                    maxPreTriggerSamples,
                                                    totalSamples,
                                                    autoStopOn,
                                                    downsampleRatio,
                                                    sizeOfOneBuffer)
    assert_pico_ok(status["runStreaming"])

    actualSampleInterval = sampleInterval.value
    actualSampleIntervalNs = actualSampleInterval * 1000

    print("Capturing at sample interval %s ns" % actualSampleIntervalNs)

    nextSample = 0
    autoStopOuter = False
    wasCalledBack = False

    def streaming_callback(handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):

        global plotbuffer, timesincelastupdate, updtime
        global nextSample, autoStopOuter, wasCalledBack
        wasCalledBack = True
        destEnd = nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        timesincelastupdate += noOfSamples/M.fs
        plotbuffer=np.roll(plotbuffer,int(-noOfSamples),axis=0)
        plotbuffer[-noOfSamples:] = (np.array(adc2mV(bufferAMax[startIndex:sourceEnd],rangeA,maxADC))/1000)
        if timesincelastupdate > updtime:
            linet.set_ydata(plotbuffer)
            linef.set_ydata(np.abs(np.fft.fft(plotbuffer*np.hanning(plotbuffersize),norm='ortho')))
            plt.pause(0.0001)
            timesincelastupdate = 0.0
        # if enabledA:
        #     bufferCompleteA[nextSample:destEnd] = bufferAMax[startIndex:sourceEnd]
        # if enabledB:
        #     bufferCompleteB[nextSample:destEnd] = bufferBMax[startIndex:sourceEnd]
        nextSample += noOfSamples
        if autoStop:
            autoStopOuter = True

    # Convert the python function into a C function pointer.
    cFuncPtr = ps4000.StreamingReadyType(streaming_callback)

    # Fetch data from the driver in a loop, copying it out of the registered buffers and into our complete one.
    while nextSample < totalSamples and not autoStopOuter and not stop:
        wasCalledBack = False
        status["getStreamingLastestValues"] = ps4000.ps4000GetStreamingLatestValues(chandle, cFuncPtr, None)
        if not wasCalledBack:
            # If we weren't called back by the driver, this means no data is ready. Sleep for a short while before trying
            # again.
            time.sleep(0.01)

    print("Done.")

    # # Convert ADC counts data to mV
    # if enabledA:
    #     adc2mVChAMax = adc2mV(bufferCompleteA, rangeA, maxADC)
    # if enabledB:
    #     adc2mVChBMax = adc2mV(bufferCompleteB, rangeB, maxADC)

    # # Create time data
    # time = np.linspace(0, (totalSamples - 1) * actualSampleIntervalNs, totalSamples)

    # Stop the scope
    # handle = chandle
    status["stop"] = ps4000.ps4000Stop(chandle)
    assert_pico_ok(status["stop"])

    # Disconnect the scope
    # handle = chandle
    status["close"] = ps4000.ps4000CloseUnit(chandle)
    assert_pico_ok(status["close"])

    # for i in range(len(M.in_map)):
    #     if M.in_map[i] == 1:
    #         M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChAMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)
    #     elif M.in_map[i] == 2:
    #         M.data[M.in_name[i]].raw = decimate(np.double(adc2mVChBMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)

    # if M.fs!=effective_fs/M.upsampling_factor:
    #     M.fs = effective_fs/M.upsampling_factor
    #     print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
    #     for i in range(len(M.in_map)):
    #         M.data[M.in_name[i]].fs = M.fs
    

def ps4000_run_measurement(M):
    # """
    # This function needs M to contain the following properties:
    #     - in_range : a list of strings specifying the voltage range.
    #         Possible voltage ranges are "10MV", "20MV", "50MV", "100MV",
    #         "200MV", "500MV", "1V", "2V", "5V", "10V", "20V", "50V", "100V"
    #     - upsampling_factor : upsampling factor
    #     - in_coupling : Coupling configuration of the channels.
    #         Can be "ac" or "dc"
    # """

    import time
    global nextSample, autoStopOuter, wasCalledBack

    # Buffer size fixed to 20k samples
    sizeOfOneBuffer = 100_000

    # Effective sampling frequency
    # If upsampling is > 1, the actual data acquisition is
    # performed at fs*upsampling, and then decimated to
    # the desired frequency
    effective_fs = M.fs * M.upsampling_factor

    # Sample interval is the duration between two consecutive samples
    # As it is the sampling frequency that is specified,
    # and the sampling interval can only take integer increments
    # of 1 nanoseconds, the actual sampling frequency might
    # necessitate adjustments
    si = round(1e9/effective_fs)
    if effective_fs != (1e9/si):
        effective_fs = (1e9/si)
        print("Warning : Sampling frequency fs changed to nearest possible value of "+str(effective_fs)+" Hz")
    
    print("Effective sampling frequency: "+str(effective_fs))

    numdesiredsamples = int(round(effective_fs*M.dur))
    numBuffersToCapture = int(np.ceil(numdesiredsamples/sizeOfOneBuffer))
    sampleInterval = ctypes.c_int32(si)
    totalSamples = sizeOfOneBuffer * numBuffersToCapture
    chandle = ctypes.c_int16()
    status = {}

    # Open PicoScope 4000 Series device
    # Returns handle to chandle for use in future API functions
    status["openunit"] = ps4000.ps4000OpenUnit(ctypes.byref(chandle))
    assert_pico_ok(status["openunit"])

    # Setup channel A
    indA = findindex(M.in_map,1)
    if indA!=None:
        enabledA = True
        rangeA = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indA]]
        if M.in_coupling[indA].capitalize()=='Dc':
           couplingA = 1
        elif M.in_coupling[indA].capitalize()=='Ac':
            couplingA = 0
        else:
            print("Input A coupling not recognized, set to 'dc'")
            couplingA = 1
            M.in_coupling[indA]='dc'
        # Create buffers ready for assigning pointers for data collection
        bufferAMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteA = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel A: enabled with range '+'PS4000_'+M.in_range[indA]+' ('+str(rangeA)+')')
    else:
        enabledA = False
        rangeA = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel A: disabled')

    # Set up channel A
    channel_range = ps4000.PS4000_RANGE['PS4000_50MV']
    status["setChA"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                            int(enabledA),
                                            couplingA,
                                            rangeA)
    assert_pico_ok(status["setChA"])

    # Setup channel B
    indB = findindex(M.in_map,2)
    if indB!=None:
        enabledB = True
        rangeB = ps4000.PS4000_RANGE['PS4000_'+M.in_range[indB]]
        # Create buffers ready for assigning pointers for data collection
        bufferBMax = np.zeros(shape=sizeOfOneBuffer, dtype=np.int16)
        # Total buffer
        bufferCompleteB = np.zeros(shape=totalSamples, dtype=np.int16)
        print('Channel B: enabled with range '+'PS4000_'+M.in_range[indB]+' ('+str(rangeB)+')')
        if M.in_coupling[indB].capitalize()=='Dc':
           couplingB = 1
        elif M.in_coupling[indB].capitalize()=='Ac':
            couplingB = 0
        else:
            print("Input B coupling not recognized, set to 'dc'")
            couplingB = 1
            M.in_coupling[indB]='dc'
    else:
        enabledB = False
        rangeB = ps4000.PS4000_RANGE['PS4000_10V']
        print('Channel B: disabled')

    # Set up channel B
    status["setChB"] = ps4000.ps4000SetChannel(chandle,
                                            ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                            int(enabledB),
                                            couplingB,
                                            rangeB)
    assert_pico_ok(status["setChB"])

    # Set data buffer location for data collection from channel A
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_A = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledA:
        status["setDataBuffersA"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_A'],
                                                        bufferAMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersA"])

    # Set data buffer location for data collection from channel B
    # Parameters :  handle = chandle
    #               source = PS4000_CHANNEL_B = 0
    #               pointer to buffer max = ctypes.byref(bufferAMax)
    #               pointer to buffer min = ctypes.byref(bufferAMin)
    #               buffer length = maxSamples
    #               segment index = 0
    #               ratio mode = PS4000_RATIO_MODE_NONE = 0
    if enabledB:
        status["setDataBuffersB"] = ps4000.ps4000SetDataBuffers(chandle,
                                                        ps4000.PS4000_CHANNEL['PS4000_CHANNEL_B'],
                                                        bufferBMax.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)),
                                                        None,
                                                        sizeOfOneBuffer)
        assert_pico_ok(status["setDataBuffersB"])

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Begin streaming mode:
    sampleUnits = ps4000.PS4000_TIME_UNITS['PS4000_NS']
    # We are not triggering:
    maxPreTriggerSamples = 0
    autoStopOn = 1
    # No downsampling:
    downsampleRatio = 1
    status["runStreaming"] = ps4000.ps4000RunStreaming(chandle,
                                                    ctypes.byref(sampleInterval),
                                                    sampleUnits,
                                                    maxPreTriggerSamples,
                                                    totalSamples,
                                                    autoStopOn,
                                                    downsampleRatio,
                                                    sizeOfOneBuffer)
    assert_pico_ok(status["runStreaming"])

    actualSampleInterval = sampleInterval.value
    actualSampleIntervalNs = actualSampleInterval * 1000

    print("Capturing at sample interval %s ns" % actualSampleIntervalNs)

    nextSample = 0
    autoStopOuter = False
    wasCalledBack = False

    def streaming_callback(handle, noOfSamples, startIndex, overflow, triggerAt, triggered, autoStop, param):
        global nextSample, autoStopOuter, wasCalledBack
        wasCalledBack = True
        destEnd = nextSample + noOfSamples
        sourceEnd = startIndex + noOfSamples
        if enabledA:
            bufferCompleteA[nextSample:destEnd] = bufferAMax[startIndex:sourceEnd]
        if enabledB:
            bufferCompleteB[nextSample:destEnd] = bufferBMax[startIndex:sourceEnd]
        nextSample += noOfSamples
        if autoStop:
            autoStopOuter = True
            print("auto")

    # if M.sig_gen:
    #     res=ps4000.ps4000SetSigGenBuiltIn(chandle,
    #             int(M.offset*1e6),      # offset voltage in uV
    #             int(M.amp/2*1e6),       # peak-to-peak votage in uV
    #             M.wave,                 # type of waveform (0 = sine wave)
    #             M.freq_start,           # start frequency in Hz
    #             M.freq_stop,            # stop frequency in Hz
    #             M.freq_change,          # frequency change per interval in Hz
    #             M.freq_int,             # interval of frequency change in seconds
    #             M.sweep_dir,            # sweep direction (0 = up)
    #             0,
    #             1,
    #             0,
    #             0,
    #             0,
    #             0
    #         )
        
    #     print("config")
    #     assert_pico_ok(res)
    #     print("fin assert")
    # print("suite")

    # Convert the python function into a C function pointer.
    cFuncPtr = ps4000.StreamingReadyType(streaming_callback)

    # Fetch data from the driver in a loop, copying it out of the registered buffers and into our complete one.
    while nextSample < totalSamples:  #and not autoStopOuter:
        wasCalledBack = False
        status["getStreamingLastestValues"] = ps4000.ps4000GetStreamingLatestValues(chandle, cFuncPtr, None)
        if not wasCalledBack:
            # If we weren't called back by the driver, this means no data is ready. Sleep for a short while before trying
            # again.
            time.sleep(0.01)

    print("Done.")

    # Find maximum ADC count value
    # handle = chandle
    # pointer to value = ctypes.byref(maxADC)
    maxADC = ctypes.c_int16(32767)

    # Convert ADC counts data to mV
    if enabledA:
        adc2mVChAMax = adc2mV(bufferCompleteA, rangeA, maxADC)
    if enabledB:
        adc2mVChBMax = adc2mV(bufferCompleteB, rangeB, maxADC)

    # Create time data
    time = np.linspace(0, (totalSamples - 1) * actualSampleIntervalNs, totalSamples)

    # Stop the scope
    # handle = chandle
    status["stop"] = ps4000.ps4000Stop(chandle)
    assert_pico_ok(status["stop"])

    # Disconnect the scope
    # handle = chandle
    status["close"] = ps4000.ps4000CloseUnit(chandle)
    assert_pico_ok(status["close"])

    for i in range(len(M.in_map)):
        if M.in_map[i] == 1:
            M.in_sig[i].raw = decimate(np.double(adc2mVChAMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)
        elif M.in_map[i] == 2:
            M.in_sig[i].raw = decimate(np.double(adc2mVChBMax[0:round(M.dur*effective_fs)])/1000,M.upsampling_factor)

    if M.fs!=effective_fs/M.upsampling_factor:
        M.fs = effective_fs/M.upsampling_factor
        print('Warning : Sampling frequency fs changed to nearest possible value of '+str(M.fs)+' Hz')
        for i,s in enumerate(M.in_sig):
            s.fs = M.fs
    


# measpy/_tools.py
#
# ----------------------------
# Utilities for measpy package
# ----------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy

import csv
import numpy as np
import h5py

def csv_to_dict(filename):
    """ Conversion from a CSV (produced by the class Measurement) to a dict
          Default separator is (,)
          First row is the key string
          The value is a list
    """
    dd={}
    with open(filename, 'r') as file:
        reader = csv.reader(file)
        for row in reader:
            dd[row[0]]=row[1:]
    return dd

def convl(fun,xx):
    if type(xx) in [np.ndarray ,list]:
        yy=list(map(fun,xx))
    else:
        yy=fun(xx) 
    return yy

def convl1(fun,xx):
    if type(xx) in [np.ndarray ,list]:
        yy=fun(xx[0])
    else:
        yy=fun(xx) 
    return yy

def add_step(a,b):
    return a+'\n -->'+b

def wrap(phase):
    """ Opposite of np.unwrap   
    """
    return np.mod((phase + np.pi), (2 * np.pi)) - np.pi

def unwrap_around_index(phase,n):
    """ Unwraps a phase array around a specified index  
    """
    return np.hstack((np.unwrap(phase[n-1::-1])[::-1],np.unwrap(phase[n:])))

def smooth(in_array,l=20):
    ker = np.ones(l)/l
    return np.convolve(in_array,ker,mode='same')

def nth_octave_bands(n,fmin=5,fmax=20000):
    """ 1/nth octave band frequency range calculation """
    nmin = int(np.ceil(n*np.log2(fmin*10**-3)))
    nmax = int(np.ceil(n*np.log2(fmax*10**-3)))
    indices = range(nmin,nmax+1)
    f_centre = 1000 * (2**(np.array(indices)/n))
    f2 = 2**(1/n/2)
    f_upper = f_centre * f2
    f_lower = f_centre / f2
    return f_centre, f_lower, f_upper

def picv(long):
    """ Create a 1D-array of length long with a unitary peak in the middle """
    return np.hstack((np.zeros(long),1,np.zeros(long-1)))

def _create_time1(fs,dur):
    return np.linspace(0,dur-1/fs,int(round(dur*fs)))  # time axis

def _create_time2(fs,length):
    return np.linspace(0,(length-1)/fs,length)  # time axis

def create_time(fs,dur=None,length=None):
    if dur==None and length==None:
        raise Exception('dur=duration in s or length=number of samples must be specified.')
    if dur!=None and length!=None:
        raise Exception("dur and length can't be both specified.")
    if dur!=None:
        return _create_time1(fs,dur)
    else:
        return _create_time2(fs,length)

def apply_fades(s,fades):
    if fades[0]>0:
        s[0:fades[0]] = s[0:fades[0]] * ((-np.cos(np.arange(fades[0])/fades[0]*np.pi)+1) / 2)
    if fades[1]>0:
        s[-fades[1]:] = s[-fades[1]:] *  ((np.cos(np.arange(fades[1])/fades[1]*np.pi)+1) / 2)
    return s

def noise(fs, dur, out_amp, freq_min, freq_max):
    """ Create band-limited noise """
    leng = int(dur*fs)
    lengs2 = int(np.ceil(leng/2))
    f = fs*np.arange(lengs2+1, dtype=float)/leng
    amp = ((f > freq_min) & (f < freq_max))*np.sqrt(leng)
    phase = 2*np.pi*(np.random.rand(lengs2+1)-0.5)
    fftx = amp*np.exp(1j*phase)
    s = out_amp*np.fft.irfft(fftx, leng)
    return s

def log_sweep(fs, dur, out_amp, freq_min, freq_max):
    """ Create log sweep """
    L = (dur-1/fs)/np.log(freq_max/freq_min)
    t = create_time(fs, dur=dur)
    s = np.sin(2*np.pi*freq_min*L*np.exp(t/L))
    return out_amp*s

def sine(fs, dur, out_amp, freq):
    s = out_amp*np.sin(2*np.pi*create_time(fs=fs, dur=dur)*freq)
    return (s)

def saw(fs, dur, out_amp, freq):
    return out_amp*wrap(2*np.pi*freq*create_time(fs,dur=dur))/np.pi
    
def tri(fs, dur, out_amp, freq):
    return out_amp*2*np.abs(wrap(2*np.pi*freq*create_time(fs,dur=dur))/np.pi)-1

def t_min(siglist):
    return min(s.t0 for s in siglist)

def t_max(siglist):
    return max(s.t0+s.dur for s in siglist)

def calc_dur_siglist(siglist):
    return t_max(siglist)-t_min(siglist)

def siglist_to_array(siglist):
    durtot = calc_dur_siglist(siglist)
    out = np.zeros((round(durtot*siglist[0].fs),len(siglist)))
    t0s = t_min(siglist)
    for i,s in enumerate(siglist):
        out[round((s.t0-t0s)*s.fs):round(((s.t0-t0s)+s.dur)*s.fs),i] = s.raw
    return out

def get_index(array,value):
    """
    Get the index of the nearest value
    """
    return np.argmin((array-value)**2)

def h5file_write_from_queue(queue, filename, dataset_name):
    """
    Write data received from a queue in HDF5 dataset

    Parameters
    ----------
    queue : queue.Queue
        A Queue which contains lists of data.
    filename : str,Path
        Path of the hdf5 file, it should already exist with an empty extensible dataset.
    dataset_name : str
        Name of the hdf5 dataset where data will be written.

    Returns
    -------
    None.

    """
    print(f"Starting saving data in {filename}/{dataset_name}")
    with h5py.File(filename, "r+") as H5file:
        item = np.asarray(queue.get())
        #Get dimension of item for multichannel case
        Nchannel, getsize = getdimension(item)
        dataset = H5file[dataset_name]
        #Get the chunksize and datatype of the dataset
        chunksize = dataset.chunks[0]
        datatype = dataset.dtype
        #Define a buffer with chuncksize and datatype
        writebuffer = np.empty((chunksize, Nchannel),dtype=datatype).squeeze()
        Npoints = getsize(item)
        buffer_position = _add_item(writebuffer, 0, item, Npoints, dataset, chunksize)
        while (item := queue.get(timeout=5)) is not None:
            item = np.asarray(item)
            Npoints = getsize(item)
            buffer_position = _add_item(
                writebuffer, buffer_position, item, Npoints, dataset, chunksize
            )
        if buffer_position > 0:
            _add_N_data(dataset, writebuffer, buffer_position)


def _add_item(writebuffer, buffer_position, item, Npoints, dataset, chunksize):
    """
    Add new item into buffer and into dataset if it fill up the buffer
    Parameters
    ----------
    writebuffer : np.array
        Data buffer.
    buffer_position : int
        Current position in the buffer (last written data+1).
    item : np.array
        New data.
    Npoints : int
        Number of new data point.
    dataset : HDF5 dataset
        Where to write data.
    chunksize : int
        Size of the dataset chunk.

    Returns
    -------
    Nrest : int
        New position in the buffer (last written data+1).

    """
    #Calcul the number of chunk to write and the new position in buffer
    Nchuncktowrite, Nrest = divmod((buffer_position + Npoints), chunksize)
    old_buffer_position = buffer_position
    #Loop over the number of new chunk in item
    for i in range(Nchuncktowrite):
        #Write item data into buffer until it is full
        writebuffer[buffer_position:] = item[
            ...,
            i * chunksize
            - np.sign(i) * old_buffer_position : (i + 1) * chunksize
            - old_buffer_position,
        ].transpose()
        #Write the whole buffer into dataset
        _add_N_data(dataset, writebuffer, chunksize)
        buffer_position = 0
    if Nrest:
        #Write Nrest data into the buffer
        writebuffer[buffer_position:Nrest] = item[
            ..., (buffer_position - Nrest) :
        ].transpose()
    return Nrest


def _add_N_data(dataset, data, N):
    #Write N data point into the dataset
    chunk_start = dataset.shape[0]
    dataset.resize(chunk_start + N, axis=0)
    dataset[chunk_start:] = data[:N]


def getdimension(item):
    if item.ndim < 2:
        return 1, np.size
    else:
        return item.shape[0], lambda item: item.shape[1]

# def _tfe_farina(y, fs, freqs):
#     """ Transfer function estimate
#         Farina's method """
#     leng = int(2**np.ceil(np.log2(len(y))))
#     Y = np.fft.rfft(y, leng)/fs
#     f = np.linspace(0, fs/2, num=round(leng/2)+1)  # frequency axis
#     L = len(y)/fs/np.log(freqs[1]/freqs[0])
#     S = 2*np.sqrt(f/L)*np.exp(-1j*2*np.pi*f*L *
#                               (1-np.log(f/freqs[0])) + 1j*np.pi/4)
#     S[0] = 0j
#     H = Y*S
#     return f, H

# def _tfe_welch(x, y, **kwargs):
#     """ Transfer function estimate (Welch's method)       
#         Arguments and defaults :
#         NFFT=None,
#         Fs=None,
#         detrend=None,
#         window=None,
#         noverlap=None,
#         pad_to=None,
#         sides=None,
#         scale_by_freq=None
#     """
#     if type(x) != type(y):
#         raise Exception(
#             'x and y must have the same type (numpy array or Signal object).')

#     # Set default values for welch's kwargs
#     if not "fs" in kwargs:
#         kwargs["fs"] = x.fs
#     if not "nperseg" in kwargs:
#         kwargs["nperseg"] = 2**(np.ceil(np.log2(x.fs)))

#     if type(x) == Signal:
#         f, p = welch(x.values_in_unit, **kwargs)
#         f, c = csd(y.values_in_unit, x.values_in_unit, **kwargs)
#         out = Spectral(desc='Transfer function between '+x.desc+' and '+y.desc,
#                        fs=x.fs,
#                        unit=y.unit+'/'+x.unit)
#         out.values = c/p
#         return out
#     else:
#         f, p = welch(x, **kwargs)
#         f, c = csd(y, x, **kwargs)
#     return f, c/p



# measpy/audio.py
#
# -----------------------------------
# Data acquisition with audio devices
# -----------------------------------
#
# Part of measpy package for signal acquisition and processing
# (c) OD - 2021 - 2023
# https://github.com/odoare/measpy


import sounddevice as sd

from ._tools import siglist_to_array, t_min

import numpy as np
from numpy.matlib import repmat

from datetime import datetime
from time import time, sleep

def audio_run_measurement(M):
    """
    Runs a measurement defined in the object of
    the class measpy.measurement.Measurement given
    as argument.

    Once the data acquisition process is terminated,
    the measurement object given in argument contains
    a property in_sig consisting of a list of signals.

    :param M: The measurement object that defines the measurement properties

    :type M: measpy.measurement.Measurement

    :return: Nothing, the measurement passed as argument is modified in place.

    """

    if M.device_type!='audio':
        print("Warning: deviceType != 'audio'. Changing to 'audio'.")
        M.device_type='audio'
    if M.in_device=='':
        print("Warning: no device specified, changing to None")
        M.in_device=None
    if M.out_sig!=None:
        if M.out_device=='':
            print("Warning: no device specified, changing to None")
            M.out_device=None

    now = datetime.now()
    M.date = now.strftime("%Y-%m-%d")
    M.time = now.strftime("%H:%M:%S")

    # Set the audio devices to use
    # And prepare the output arrays
    if M.out_sig!=None:
        outx = siglist_to_array(M.out_sig)
        tmin = t_min(M.out_sig)
        if M.in_sig!=None:
            sd.default.device=(M.in_device,M.out_device)
        else:
            sd.default.device=(M.out_device)
    else:
        tmin = 0
        if M.in_sig!=None:
            sd.default.device=(M.in_device)
        else:
            raise Exception('No input nor output defined.')

    if M.out_sig==None:
        y = sd.rec(int(M.dur * M.fs),
                samplerate=M.fs,
                mapping=M.in_map,
                blocking=False)
    else:
        if M.in_sig==None:
            y = sd.play(outx,
                    samplerate=M.fs,
                    mapping=M.out_map,
                    blocking=False)
        else:
            y = sd.playrec(outx,
                    samplerate=M.fs,
                    input_mapping=M.in_map,
                    output_mapping=M.out_map,
                    blocking=False)
            
    sd.wait()

    if M.in_sig!=None:
        print(M.in_sig)
        for i,s in enumerate(M.in_sig):
            s.raw = np.array(y[:,i])
            s.t0 = tmin

def audio_run_synced_measurement(M,in_chan=0,out_chan=0):
    """
    Before running a measurement, one second of silence
    is added at the begining and end of the selected output channel.
    The measurement is then run, and the time lag between
    a selected acquired signal and the output signal is computed
    from cross-correlation calculation.
    All the acquired signals are then re-synced from the time lag value.

    :param M: The measurement object
    :type M: measpy.measurement.Measurement
    :param out_chan: The selected output channel for synchronization. It is the index of the selected output signal in the list ``M.out_sig``
    :type out_chan: int
    :param in_chan: The selected input channel for synchronization. It is the index of the selected input signal in the list ``M.in_sig``
    :type in_chan: int

    """
    M.sync_prepare(out_chan=out_chan)
    audio_run_measurement(M)
    M.sync_render(out_chan=out_chan,in_chan=in_chan)
    return M

def audio_get_devices():
   """
   Returns a list of audio devices present in the system
   """
   return sd.query_devices()
 